import contextlib
import re

import pytest

import outlines.generate as generate
import outlines.models as models
import outlines.samplers as samplers

##########################################
# Model Fixtures
##########################################


@pytest.fixture(scope="session")
def model_llamacpp(tmp_path_factory):
    return models.llamacpp(
        repo_id="M4-ai/TinyMistral-248M-v2-Instruct-GGUF",
        filename="TinyMistral-248M-v2-Instruct.Q4_K_M.gguf",
    )


@pytest.fixture(scope="session")
def model_exllamav2(tmp_path_factory):
    from huggingface_hub import snapshot_download

    tmp_dir = tmp_path_factory.mktemp("model_download")
    model_path = snapshot_download(
        repo_id="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4.6-exl2",
        cache_dir=tmp_dir,
    )

    return models.exl2(
        model_path=model_path,
        cache_q4=True,
        paged=False,
    )


@pytest.fixture(scope="session")
def model_mlxlm(tmp_path_factory):
    return models.mlxlm("mlx-community/TinyLlama-1.1B-Chat-v1.0-4bit")


@pytest.fixture(scope="session")
def model_mlxlm_phi3(tmp_path_factory):
    return models.mlxlm("mlx-community/Phi-3-mini-4k-instruct-4bit")


@pytest.fixture(scope="session")
def model_transformers_random(tmp_path_factory):
    return models.transformers("hf-internal-testing/tiny-random-gpt2", device="cpu")


@pytest.fixture(scope="session")
def model_transformers_opt125m(tmp_path_factory):
    return models.transformers("facebook/opt-125m", device="cpu")


@pytest.fixture(scope="session")
def model_mamba(tmp_path_factory):
    return models.mamba(model_name="state-spaces/mamba-130m-hf", device="cpu")


@pytest.fixture(scope="session")
def model_bart(tmp_path_factory):
    from transformers import AutoModelForSeq2SeqLM

    return models.transformers(
        "facebook/bart-base", device="cpu", model_class=AutoModelForSeq2SeqLM
    )


@pytest.fixture(scope="session")
def model_transformers_vision(tmp_path_factory):
    import torch
    from transformers import LlavaNextForConditionalGeneration

    return models.transformers_vision(
        "llava-hf/llava-v1.6-mistral-7b-hf",
        model_class=LlavaNextForConditionalGeneration,
        device="cuda",
        model_kwargs=dict(
            torch_dtype=torch.bfloat16,
            load_in_4bit=True,
            low_cpu_mem_usage=True,
        ),
    )


@pytest.fixture(scope="session")
def model_vllm(tmp_path_factory):
    return models.vllm("facebook/opt-125m", gpu_memory_utilization=0.1)


# TODO: exllamav2 failing in main, address in https://github.com/dottxt-ai/outlines/issues/808
# TODO: t5 tokenizer doesn't work with streaming
"""
@pytest.fixture(scope="session")
def model_exllamav2(tmp_path_factory):
    return models.exllamav2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        device="cpu"
    )

@pytest.fixture(scope="session")
def model_t5(tmp_path_factory):
    from transformers import AutoModelForSeq2SeqLM

    return models.transformers(
        "EleutherAI/pile-t5-base", device="cpu", model_class=AutoModelForSeq2SeqLM
    )
"""


ALL_MODEL_FIXTURES = (
    "model_llamacpp",
    "model_exllamav2",
    "model_mlxlm",
    "model_mlxlm_phi3",
    "model_transformers_random",
    "model_transformers_opt125m",
    "model_mamba",
    "model_bart",
    "model_transformers_vision",
    "model_vllm",
)


##########################################
# Stuctured Generation Inputs
##########################################


@pytest.fixture()
def sample_schema():
    from pydantic import BaseModel, conint, conlist, constr

    class SampleSchema(BaseModel):
        title: constr(max_length=10)
        numbers: conlist(conint(strict=True), min_length=3, max_length=3)
        labels: conlist(constr(min_length=1, max_length=5), min_length=3, max_length=3)

    return SampleSchema


@pytest.fixture()
def sample_choices():
    return ["foo", "bar", "baz"]


@pytest.fixture()
def sample_lark_grammar():
    # from https://github.com/lark-parser/lark/blob/master/docs/grammar.md
    return """
    ?start: hello_world "!" number
    hello_world: ("hello" | "world") ~ 3
    number: ("0".."9") ~ 5
    thanks: "Thank"i " for testing!"
    """


REGEX_PATTERNS = [
    "a b c d e",  # ensure proper tokenizer whitespace prefix handling
    "(123456789)|(abcdefghijklmnop)",  # ensure consistent correct sequence handling during batch
    r"([a-z]{10})@([a-z]{5})\.([a-z]{3})",  # email example
]


###########################################
# Model/Generator Pair Behavior Definitions
###########################################


def enforce_not_implemented(model_fixture, *task_names):
    """
    Per `NOT_IMPLEMENTED`, mapping, if a model hasn't implemented a task,
    assert an NotImplementedError is raised. Otherwise, run normally
    """
    NOT_IMPLEMENTED = {
        "stream": ["model_transformers_vision", "model_vllm"],
        "batch": ["model_llamacpp", "model_mlxlm", "model_mlxlm_phi3"],
        "beam_search": ["model_llamacpp", "model_mlxlm", "model_mlxlm_phi3"],
        "multiple_samples": ["model_llamacpp", "model_mlxlm", "model_mlxlm_phi3"],
        "cfg": ["model_llamacpp"],  # TODO: fix llama_cpp tokenizer
    }
    for task_name in task_names:
        if model_fixture in NOT_IMPLEMENTED.get(task_name, []):
            return pytest.raises(NotImplementedError)
    else:
        return contextlib.nullcontext()


def get_inputs(fixture_name, batch_size=None):
    """Get generator kwargs, just the prompt by default, but include images for transformers_visian"""
    from io import BytesIO
    from urllib.request import urlopen

    from PIL import Image  # type: ignore

    prompts = ["abcd", "efgh", "1234", "5678", "foo", "bar", "baz", "bif"]
    prompts = prompts[0] if batch_size is None else prompts[:batch_size]

    if fixture_name.endswith("_vision"):
        img_url = "https://python-pillow.org/pillow-perf/static/space_pil_lanczos.png"
        img = Image.open(BytesIO(urlopen(img_url).read())).convert("RGB")

        if batch_size is None:
            return {"prompts": f"<image> {prompts}", "media": [img]}
        else:
            return {
                "prompts": [f"<image> {p}" for p in prompts],
                "media": [[img] for _ in range(batch_size)],
            }

    else:
        return {"prompts": prompts}


###########################################
# Tests
###########################################


@pytest.mark.parametrize("sampler_name", ("greedy", "multinomial", "beam_search"))
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_text(request, model_fixture, sampler_name):
    model = request.getfixturevalue(model_fixture)
    generator = generate.text(model, getattr(samplers, sampler_name)())
    with enforce_not_implemented(model_fixture, sampler_name):
        res = generator(**get_inputs(model_fixture), max_tokens=10)
        assert isinstance(res, str)


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_regex(request, model_fixture, pattern):
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(model, pattern)
    res = generator(**get_inputs(model_fixture), max_tokens=20)
    assert re.fullmatch(pattern, res) is not None, res


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_fsm(request, model_fixture, pattern):
    import interegular

    model = request.getfixturevalue(model_fixture)
    generator = generate.fsm(model, interegular.parse_pattern(pattern).to_fsm())
    res = generator(**get_inputs(model_fixture))
    assert re.fullmatch(pattern, res) is not None, res


@pytest.mark.skip(
    "Fix issues with JSON, some models fail this test https://github.com/dottxt-ai/outlines/issues/985"
)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_json(request, model_fixture, sample_schema):
    model = request.getfixturevalue(model_fixture)
    generator = generate.json(model, sample_schema)
    # asserts valid within call
    generator(**get_inputs(model_fixture), max_tokens=100)


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_choice(request, model_fixture, sample_choices):
    model = request.getfixturevalue(model_fixture)
    generator = generate.choice(model, sample_choices)
    res = generator(**get_inputs(model_fixture))
    assert res in sample_choices


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_choice_twice(request, model_fixture, sample_choices):
    model = request.getfixturevalue(model_fixture)
    generator = generate.choice(model, sample_choices)
    res = generator(**get_inputs(model_fixture))
    assert res in sample_choices
    res = generator(**get_inputs(model_fixture))
    assert res in sample_choices


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_format_bool(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    generator = generate.format(model, bool)
    res = generator(**get_inputs(model_fixture))
    assert isinstance(res, bool)


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_cfg(request, model_fixture, sample_lark_grammar):
    from lark import Lark

    from outlines import grammars

    model = request.getfixturevalue(model_fixture)
    with enforce_not_implemented(model_fixture, "cfg"):
        generator = generate.cfg(model, sample_lark_grammar)
        res = generator(**get_inputs(model_fixture))
        # validate legal with the grammar via lark
        # TODO: cleanup PartialLark so doesn't modify Lark globally
        import importlib

        import lark.lark

        importlib.reload(lark.lark)
        Lark(
            sample_lark_grammar, parser="lalr", import_paths=[grammars.GRAMMAR_PATH]
        ).parse(res)


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_text_stream(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    generator = generate.text(model)
    with enforce_not_implemented(model_fixture, "stream"):
        for token in generator.stream(**get_inputs(model_fixture), max_tokens=10):
            assert isinstance(token, str)


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_regex_stream(request, model_fixture, pattern):
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(model, pattern)
    with enforce_not_implemented(model_fixture, "stream"):
        output = ""
        for token in generator.stream(**get_inputs(model_fixture), max_tokens=20):
            output += token
        assert re.fullmatch(pattern, output) is not None, output


@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_batch_text(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    generator = generate.text(model)
    with enforce_not_implemented(model_fixture, "batch"):
        res = generator(**get_inputs(model_fixture, 2), max_tokens=10)
        assert isinstance(res, list)
        assert isinstance(res[0], str)


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_regex_batch(request, model_fixture, pattern):
    """Ensure batch requests work and fsm order is maintained"""
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(model, pattern)
    with enforce_not_implemented(model_fixture, "batch"):
        outputs = generator(**get_inputs(model_fixture, 4), max_tokens=20)
        for output in outputs:
            assert re.fullmatch(pattern, output) is not None, output


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_regex_batch_stream(request, model_fixture, pattern):
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(model, pattern)
    with enforce_not_implemented(model_fixture, "batch", "stream"):
        outputs = ["", ""]
        for tokens in generator.stream(**get_inputs(model_fixture, 2), max_tokens=20):
            outputs[0] += tokens[0]
            outputs[1] += tokens[1]
        for output in outputs:
            assert re.fullmatch(pattern, output) is not None, output


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
def test_generate_regex_single_multinomial(request, model_fixture, pattern):
    """Ensure batch requests work and fsm order is maintained"""
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(model, pattern, sampler=samplers.multinomial(4))
    with enforce_not_implemented(model_fixture, "multiple_samples"):
        output_sample_groups = generator(**get_inputs(model_fixture), max_tokens=40)
        for output in output_sample_groups:
            assert re.fullmatch(pattern, output) is not None, output


@pytest.mark.parametrize("pattern", REGEX_PATTERNS)
@pytest.mark.parametrize("model_fixture", ALL_MODEL_FIXTURES)
@pytest.mark.parametrize("sampler_name", ("multinomial", "beam_search"))
def test_generate_regex_batch_multi_sample(
    request, model_fixture, pattern, sampler_name
):
    """Ensure batch requests work and fsm order is maintained"""
    model = request.getfixturevalue(model_fixture)
    generator = generate.regex(
        model, pattern, sampler=getattr(samplers, sampler_name)(4)
    )
    with enforce_not_implemented(model_fixture, "batch", "multiple_samples"):
        output_batch_groups = generator(**get_inputs(model_fixture, 4), max_tokens=40)
        for output_sample_groups in output_batch_groups:
            for output in output_sample_groups:
                assert re.fullmatch(pattern, output) is not None, output



import importlib
from copy import copy

import lark.lark
import pytest
from lark.indenter import DedentError
from lark.lexer import UnexpectedCharacters, UnexpectedToken

from outlines.fsm.parsing import PartialLark, PartialPythonIndenter


@pytest.fixture
def cleanup_lark_import():
    yield
    # Clean up lark.lark.LarkOptions._defaults
    importlib.reload(lark.lark)


def test_partial_parsing(cleanup_lark_import):
    lp = PartialLark.open_from_package(
        "tests",
        "partial_python.lark",
        ["fsm"],
        parser="lalr",
        postlex=PartialPythonIndenter(),
        start="file_input",
        deterministic=True,
    )

    # End with a potentially unfinished NAME
    parser_state = lp.parse("x")
    assert parser_state.state_stack == [0]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 15)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    # End with an ignored token
    parser_state = lp.parse("x ")
    assert parser_state.state_stack == [0, 692]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 1)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    # Could be a complete `=` or the start of a `==`
    parser_state = lp.parse("x =")
    assert parser_state.state_stack == [0, 692]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert any(
        term_info.terminal_name == "EQUAL"
        for term_info in last_token.value.terminals_and_info
    )
    assert not parser_state.value_stack

    parser_state = lp.parse("x = '")
    assert parser_state.state_stack == [0, 58, 59]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 6)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    parser_state = lp.parse("x = 'hi")
    assert parser_state.state_stack == [0, 58, 59]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 6, 6, 6)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    parser_state = lp.parse("x = ('hi")
    assert parser_state.state_stack == [0, 58, 59, 254]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 6, 6, 6)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    parser_state = lp.parse("def")
    assert parser_state.state_stack == [0]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"
    assert last_token.value.fsm_state_seq == (0, 26, 99, 100)
    assert last_token.value.is_not_finished is True
    assert not parser_state.value_stack

    # Now, try something incremental
    last_lexer_state = parser_state.lexer.state
    last_lexer_state.text += " blah()"
    lp.parse_from_state(parser_state, is_end=False)
    last_token = parser_state.lexer.state.last_token
    assert not parser_state.value_stack

    last_lexer_state = parser_state.lexer.state
    last_valid_token = last_lexer_state.last_token
    assert last_valid_token.type == "RPAR"
    assert not parser_state.value_stack

    # Something incremental and a little more complicated
    parser_state = lp.parse("x = 1\ndef foo(x):\n  ")
    assert parser_state.state_stack == [0, 94, 600, 601, 602, 607, 608, 269]
    last_lexer_state = parser_state.lexer.state
    last_lexer_state.text += "  return x"

    lp.parse_from_state(parser_state, is_end=False)
    assert parser_state.state_stack == [
        0,
        94,
        600,
        601,
        602,
        607,
        608,
        269,
        764,
        95,
        305,
    ]
    last_token = parser_state.lexer.state.last_token
    assert last_token.type == "partial"

    with pytest.raises(UnexpectedToken):
        lp.parse("def \n")

    with pytest.raises(UnexpectedToken):
        lp.parse("def hot no")

    lp = PartialLark.open_from_package(
        "tests",
        "partial_python.lark",
        ["fsm"],
        parser="lalr",
        postlex=PartialPythonIndenter(),
        start="file_input",
        use_value_stack=True,
    )
    parser_state = lp.parse("x = ('hi")
    lp.parse_from_state(parser_state, is_end=False)
    assert len(parser_state.state_stack) == 4
    assert parser_state.value_stack[-1].type == "LPAR"


def test_sequential_parse_example(cleanup_lark_import):
    input_tokens = [
        "x ",
        "= ",
        "1",
        "\nde",
        "f ",
        "foo(",
        "x)",
        ":\n",
        "  ",
        "  return x",
        " + 1",
        "\n",
        "z ",
        "= ",
        "foo(",
        '"hi',
        '")\n',
    ]
    vocab = sorted(set(input_tokens))

    lp = PartialLark.open_from_package(
        "tests",
        "partial_python.lark",
        ["fsm"],
        parser="lalr",
        postlex=PartialPythonIndenter(),
        start="file_input",
        deterministic=True,
    )
    parser_state = lp.parse("")

    token_seq = ""
    for i, token in enumerate(input_tokens):
        token_seq += token

        lex_state = parser_state.lexer.state
        lex_state.text = token_seq

        lp.parse_from_state(parser_state, is_end=False)

        next_vocab = set()
        for test_token in vocab:
            ps = copy(parser_state)
            ls = ps.lexer.state
            ls.text = token_seq + test_token

            if i + 1 < len(input_tokens) and test_token == input_tokens[i + 1]:
                lp.parse_from_state(ps, is_end=False)
                next_vocab.add(test_token)
            else:
                try:
                    lp.parse_from_state(ps, is_end=False)
                    next_vocab.add(test_token)
                except (EOFError, UnexpectedToken, UnexpectedCharacters, DedentError):
                    pass

        if i + 1 == len(input_tokens):
            assert all(tk in next_vocab for tk in ["\n", "\nde", "  ", " + 1"])


# TODO: Remove once fsm_union and walk_fsm are implemented in Outlines-Core
import interegular  # noqa

from outlines.fsm.parsing import fsm_union, walk_fsm  # noqa


def test_outlines_interegular_union_consistency():
    fsm0 = interegular.parse_pattern(r"abc").to_fsm()
    fsm1 = interegular.parse_pattern(r"WXYZ").to_fsm()
    fsm2 = interegular.parse_pattern(r"12345").to_fsm()

    interegular_unioned_fsm = fsm0 | fsm1 | fsm2
    outlines_unioned_fsm, _ = fsm_union([fsm0, fsm1, fsm2])

    assert list(outlines_unioned_fsm.strings()) == list(
        interegular_unioned_fsm.strings()
    )


def _reconstruct_fsms(fsm, fsms_to_trans_finals):
    """Reconstruct the original fsms for testing purposes"""
    reconstructed_fsms = []
    for transitions, finals, state_map in fsms_to_trans_finals.values():
        inv_state_map = {new: orig for orig, news in state_map.items() for new in news}
        states = set(inv_state_map.values())
        initial = inv_state_map.get(fsm.initial) or next(
            (orig for orig, news in state_map.items() if fsm.initial in news), None
        )
        finals = {inv_state_map[s] for s in finals}

        transition_map = {}
        alphabet = {}
        for trans_id, (from_state, to_state) in enumerate(transitions):
            orig_from, orig_to = inv_state_map[from_state], inv_state_map[to_state]
            # Collect symbols associated with the transition
            symbols = {
                symbol
                for trans, dest in fsm.map.get(from_state, {}).items()
                if dest == to_state
                for symbol in fsm.alphabet.by_transition.get(trans, [])
            }
            if symbols:
                # NOTE: THIS RECONSTRUCTOR DOESNT WORK FOR MORE THAN ONE TRANSITION PER SYMBOL
                assert len(symbols) == 1
                symbol = list(symbols)[0]
                alphabet[symbol] = trans_id
                transition_map.setdefault(orig_from, {})[trans_id] = orig_to

        reconstructed_fsms.append(
            interegular.fsm.FSM(
                alphabet=interegular.fsm.Alphabet(alphabet),
                states=frozenset(states),
                initial=initial,
                finals=frozenset(finals),
                map=transition_map,
                __no_validation__=True,
            )
        )
    return reconstructed_fsms


def test_fsm_to_trans_finals_reconstruction():
    """Assert that _fsms_to_trans_finals is correct by reconstructing original fsms"""
    fsm0 = interegular.parse_pattern(r"abc").to_fsm()
    fsm1 = interegular.parse_pattern(r"XYZ").to_fsm()
    fsm2 = interegular.parse_pattern(r"12345").to_fsm()

    fsm, _fsms_to_trans_finals = fsm_union([fsm0, fsm1, fsm2])

    reconstructed = _reconstruct_fsms(fsm, _fsms_to_trans_finals)

    # assert reconstruction equivalent
    assert list(fsm0.strings()) == list(reconstructed[0].strings())
    assert list(fsm1.strings()) == list(reconstructed[1].strings())
    assert list(fsm2.strings()) == list(reconstructed[2].strings())


def test_walk_fsm():
    fsm = interegular.parse_pattern(r"abc*d").to_fsm()
    # convert to BetterFSM
    fsm = fsm_union([fsm])[0]

    # if match, produce equivalent number of states, assert state can terminate
    transitions = [fsm.alphabet[letter] for letter in "abcccd"]
    accepted_states = walk_fsm(fsm, transitions, fsm.initial, full_match=True)
    assert len(accepted_states) == len(transitions)
    assert accepted_states[-1] in fsm.finals

    # if no match, assert empty
    accepted_states = walk_fsm(
        fsm, [fsm.alphabet[letter] for letter in "b"], fsm.initial, full_match=True
    )
    assert accepted_states == []

    # if full_match, but last state not present, assert empty
    accepted_states = walk_fsm(
        fsm, [fsm.alphabet[letter] for letter in "abc"], fsm.initial, full_match=True
    )
    assert accepted_states == []



from pathlib import Path

GRAMMAR_PATH = Path(__file__).parent / "grammars"


def read_grammar(grammar_file_name, base_grammar_path=GRAMMAR_PATH):
    """Read grammar file from default grammar path"""
    full_path = base_grammar_path / grammar_file_name
    with open(full_path) as file:
        return file.read()


arithmetic = read_grammar("arithmetic.lark")
json = read_grammar("json.lark")



from io import BytesIO
from urllib.request import urlopen

import pytest
from PIL import Image
from transformers import AutoProcessor, LlavaForConditionalGeneration

import outlines
from outlines.models.transformers_vision import transformers_vision

IMAGE_URLS = [
    "https://upload.wikimedia.org/wikipedia/commons/2/25/Siam_lilacpoint.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/7/71/2010-kodiak-bear-1.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/b/be/Tamias-rufus-001.jpg",
]


def img_from_url(url):
    img_byte_stream = BytesIO(urlopen(url).read())
    return Image.open(img_byte_stream).convert("RGB")


@pytest.fixture(scope="session")
def model(tmp_path_factory):
    return transformers_vision(
        "trl-internal-testing/tiny-random-LlavaForConditionalGeneration",
        model_class=LlavaForConditionalGeneration,
        device="cpu",
    )


@pytest.fixture(scope="session")
def processor(tmp_path_factory):
    return AutoProcessor.from_pretrained("llava-hf/llava-interleave-qwen-0.5b-hf")


def test_single_image_text_gen(model, processor):
    conversation = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "What is this?"}, {"type": "image"}],
        },
    ]
    generator = outlines.generate.text(model)
    sequence = generator(
        processor.apply_chat_template(conversation),
        [img_from_url(IMAGE_URLS[0])],
        seed=10000,
        max_tokens=10,
    )
    assert isinstance(sequence, str)


def test_multi_image_text_gen(model, processor):
    """If the length of image tags and number of images we pass are > 1 and equal,
    we should yield a successful generation.
    """
    conversation = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What do all these have in common?"},
            ]
            + [{"type": "image"} for _ in range(len(IMAGE_URLS))],
        },
    ]
    generator = outlines.generate.text(model)
    sequence = generator(
        processor.apply_chat_template(conversation),
        [img_from_url(i) for i in IMAGE_URLS],
        seed=10000,
        max_tokens=10,
    )
    assert isinstance(sequence, str)


def test_mismatched_image_text_gen(model, processor):
    """If the length of image tags and number of images we pass are unequal,
    we should raise an error.
    """
    conversation = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "I'm passing 3 images, but only 1 image tag"},
                {"type": "image"},
            ],
        },
    ]
    generator = outlines.generate.text(model)
    with pytest.raises(ValueError):
        _ = generator(
            processor.apply_chat_template(conversation),
            [img_from_url(i) for i in IMAGE_URLS],
            seed=10000,
            max_tokens=10,
        )


def test_single_image_choice(model, processor):
    conversation = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "What is this?"}, {"type": "image"}],
        },
    ]
    choices = ["cat", "dog"]
    generator = outlines.generate.choice(model, choices)
    sequence = generator(
        processor.apply_chat_template(conversation),
        [img_from_url(IMAGE_URLS[0])],
        seed=10000,
        max_tokens=10,
    )
    assert isinstance(sequence, str)
    assert sequence in choices



from io import BytesIO
from urllib.request import urlopen

import pytest
from PIL import Image  # type: ignore

from outlines.generate.api import VisionSequenceGeneratorAdapter

IMG_URI = "https://upload.wikimedia.org/wikipedia/en/a/a9/Example.jpg"
PIL_IMG = Image.open(BytesIO(urlopen(IMG_URI).read())).convert("RGB")


@pytest.mark.parametrize(
    "prompts,media,type_error",
    [
        ("single prompt", [PIL_IMG], False),
        (["prompt0", "prompt1"], [[PIL_IMG], [PIL_IMG]], False),
        ("single prompt", [PIL_IMG, PIL_IMG], False),
        (["prompt0", "prompt1"], [[PIL_IMG, PIL_IMG], [PIL_IMG]], False),
        ("single prompt", "this isn't an image, it's a string", True),
        ("single prompt", PIL_IMG, True),
        (["prompt0", "prompt1"], [PIL_IMG], True),
        (["prompt0", "prompt1"], [[PIL_IMG]], True),
        (["prompt0", "prompt1"], [[[PIL_IMG]], [[PIL_IMG]]], True),
    ],
)
def test_vision_sequence_generator_validate_types(prompts, media, type_error):
    """Ensure inputs are validated correctly"""
    if type_error:
        with pytest.raises(TypeError):
            VisionSequenceGeneratorAdapter._validate_prompt_media_types(prompts, media)
    else:
        VisionSequenceGeneratorAdapter._validate_prompt_media_types(prompts, media)



"""ReAct

This example was inspired by the LQML library [1]_. The ReAct framework was
first developed in [2]_ and augments Chain-of-Thought prompting with the ability
for the model to query external sources.

References
----------
.. [1] Beurer-Kellner, L., Fischer, M., & Vechev, M. (2022). Prompting Is Programming: A Query Language For Large Language Models. arXiv preprint arXiv:2212.06094.
.. [2] Yao, S., Zhao, J., Yu, D., Du, N., Shafran, I., Narasimhan, K., & Cao, Y. (2022). React: Synergizing reasoning and acting in language models. arXiv preprint arXiv:2210.03629.

"""
import requests  # type: ignore

import outlines
import outlines.generate as generate
import outlines.models as models


@outlines.prompt
def build_reAct_prompt(question):
    """What is the elevation range for the area that the eastern sector of the Colorado orogeny extends into?
    Tho 1: I need to search Colorado orogeny, find the area that the eastern sector of the Colorado ...
    Act 2: Search 'Colorado orogeny'
    Obs 2: The Colorado orogeny was an episode of mountain building (an orogeny) ...
    Tho 3: It does not mention the eastern sector. So I need to look up eastern sector.
    ...
    Tho 4: High Plains rise in elevation from around 1,800 to 7,000 ft, so the answer is 1,800 to 7,000 ft.
    Act 5: Finish '1,800 to 7,000 ft'
    {{ question }}
    """


@outlines.prompt
def add_mode(i, mode, result, prompt):
    """{{ prompt }}
    {{ mode }} {{ i }}: {{ result }}
    """


def search_wikipedia(query: str):
    url = f"https://en.wikipedia.org/w/api.php?format=json&action=query&prop=extracts&exintro&explaintext&redirects=1&titles={query}&origin=*"
    response = requests.get(url)
    page = response.json()["query"]["pages"]
    return ".".join(list(page.values())[0]["extract"].split(".")[:2])


prompt = build_reAct_prompt("Where is Apple Computers headquarted? ")
model = models.openai("gpt-4o-mini")

mode_generator = generate.choice(model, choices=["Tho", "Act"])
action_generator = generate.choice(model, choices=["Search", "Finish"])
text_generator = generate.text(model)

for i in range(1, 10):
    mode = mode_generator(prompt, max_tokens=128)
    prompt = add_mode(i, mode, "", prompt)

    if mode == "Tho":
        thought = text_generator(prompt, stop_at="\n", max_tokens=128)
        prompt += f"{thought}"
    elif mode == "Act":
        action = action_generator(prompt, max_tokens=128)
        prompt += f"{action} '"

        subject = text_generator(prompt, stop_at=["'"], max_tokens=128)
        # Apple Computers headquartered
        subject = " ".join(subject.split()[:2])
        prompt += f"{subject}'"

        if action == "Search":
            result = search_wikipedia(subject)
            prompt = add_mode(i, "Obs", result, prompt)
        else:
            break

print(prompt)



import math

import pytest
import torch

from outlines.samplers import (
    BeamSearchSampler,
    GreedySampler,
    MultinomialSampler,
    beam_search,
    greedy,
    keep_top_k_logits,
    keep_top_p_logits,
    multinomial,
    rescale_logits,
)


def compute_logprobs(logits):
    return torch.nn.functional.log_softmax(logits, dim=-1)


def test_aliases():
    assert greedy == GreedySampler
    assert multinomial == MultinomialSampler
    assert beam_search == BeamSearchSampler


def test_greedy():
    sampler = GreedySampler()
    logits = torch.tensor([[1.0, 2.0, 5.0]])
    weights = torch.tensor([0])
    next_token_ids, ancestors, weights = sampler(logits, weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[2]]))
    assert ancestors.equal(torch.tensor([0]))
    assert weights.equal(logprobs[..., 2])

    sampler = GreedySampler()
    logits = torch.tensor([[10.0, 0.0, 3.0], [-math.inf, 2.0, 5.0]])
    weights = torch.tensor([0, 0])
    next_token_ids, ancestors, weights = sampler(logits, weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[0], [2]]))
    assert ancestors.equal(torch.tensor([0, 1]))
    assert weights.equal(torch.tensor([logprobs[0, 0], logprobs[1, 2]]))


def test_multinomial():
    rng = torch.Generator()
    rng.manual_seed(239)

    sampler = MultinomialSampler()
    logits = torch.tensor([[1.0, 4.0, 5.0]])
    weights = torch.tensor([0])
    next_token_ids, ancestors, weights = sampler(logits, weights, rng)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[2]]))
    assert ancestors.equal(torch.tensor([0]))
    assert weights.equal(torch.tensor([logprobs[:, 2]]))

    sampler = MultinomialSampler()
    logits = torch.tensor([[10.0, 0.0, 9.0], [-math.inf, 4.0, 5.0]])
    weights = torch.tensor([0, 0])
    next_token_ids, ancestors, weights = sampler(logits, weights, rng)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[0], [2]]))
    assert ancestors.equal(torch.tensor([0, 1]))
    assert weights.equal(torch.tensor([logprobs[0, 0], logprobs[1, 2]]))


def test_multinomial_init():
    sampler = MultinomialSampler()
    assert sampler.logits_processors == []

    sampler = MultinomialSampler(3)
    assert sampler.logits_processors == []

    sampler = MultinomialSampler(top_k=1)
    assert len(sampler.logits_processors) == 1

    sampler = MultinomialSampler(top_p=0.95)
    assert len(sampler.logits_processors) == 1

    sampler = MultinomialSampler(top_k=1, top_p=0.95)
    assert len(sampler.logits_processors) == 1

    sampler = MultinomialSampler(temperature=1.0)
    assert len(sampler.logits_processors) == 1

    sampler = MultinomialSampler(top_k=1, temperature=1.0)
    assert len(sampler.logits_processors) == 2

    sampler = MultinomialSampler(top_p=0.95, temperature=1.0)
    assert len(sampler.logits_processors) == 2


def test_top_k():
    with pytest.raises(ValueError, match="`k` must be a strictly"):
        keep_top_k_logits(-1)

    with pytest.raises(ValueError, match="`k` must be a strictly"):
        keep_top_k_logits(0.1)

    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0]])

    logits_processor = keep_top_k_logits(1)
    result = logits_processor(logits)
    assert result.equal(torch.tensor([[-math.inf, -math.inf, -math.inf, 4.0]]))

    logits_processor = keep_top_k_logits(10)
    result = logits_processor(logits)
    assert result.equal(torch.tensor([[1.0, 2.0, 3.0, 4.0]]))

    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    logits_processor = keep_top_k_logits(2)
    result = logits_processor(logits)
    assert result.equal(
        torch.tensor(
            [[-math.inf, -math.inf, 3.0, 4.0], [-math.inf, -math.inf, 7.0, 8.0]]
        )
    )


def test_top_p():
    with pytest.raises(ValueError, match="`p` must be a floating point"):
        keep_top_p_logits(-0.1)

    with pytest.raises(ValueError, match="`p` must be a floating point"):
        keep_top_p_logits(0.0)

    with pytest.raises(ValueError, match="`p` must be a floating point"):
        keep_top_p_logits(1.1)

    logits = torch.tensor([[1.0, 1.01, 1.02, 4.0]])

    logits_processor = keep_top_p_logits(0.1)
    result = logits_processor(logits)
    assert result.equal(torch.tensor([[-math.inf, -math.inf, -math.inf, 4.0]]))

    logits_processor = keep_top_p_logits(0.95)
    result = logits_processor(logits)
    assert result.equal(torch.tensor([[-math.inf, 1.01, 1.02, 4.0]]))

    logits_processor = keep_top_p_logits(1.0)
    result = logits_processor(logits)
    assert result.equal(torch.tensor([[1.0, 1.01, 1.02, 4.0]]))

    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])

    logits_processor = keep_top_p_logits(0.1)
    result = logits_processor(logits)
    assert result.equal(
        torch.tensor(
            [
                [-math.inf, -math.inf, -math.inf, 4.0],
                [-math.inf, -math.inf, -math.inf, 8.0],
            ]
        )
    )

    logits_processor = keep_top_p_logits(0.95)
    result = logits_processor(logits)
    assert result.equal(
        torch.tensor(
            [
                [-math.inf, 2.0, 3.0, 4.0],
                [-math.inf, 6.0, 7.0, 8.0],
            ]
        )
    )

    logits_processor = keep_top_p_logits(1.0)
    result = logits_processor(logits)
    assert result.equal(
        torch.tensor(
            [
                [1.0, 2.0, 3.0, 4.0],
                [5.0, 6.0, 7.0, 8.0],
            ]
        )
    )


def test_rescale():
    with pytest.raises(ValueError, match="`temperature` must"):
        rescale_logits(1)

    with pytest.raises(ValueError, match="`temperature` must"):
        rescale_logits(-0.1)

    with pytest.raises(ValueError, match="Please use the greedy sampler"):
        rescale_logits(0.0)


def test_beam_search():
    # Two beams, single sequence
    sampler = BeamSearchSampler(2)
    logits = torch.tensor([[0.0, 1.0], [2.0, 0.0]])
    init_weights = torch.tensor([0, 1.0])
    next_token_ids, ancestors, weights = sampler(logits, init_weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[0], [1]]))
    assert ancestors.equal(torch.tensor([1, 0]))
    assert weights.equal(
        torch.tensor([init_weights[1] + logprobs[1][0], logprobs[0][1]])
    )

    # Make sure that initial samples are different
    sampler = BeamSearchSampler(2)
    logits = torch.tensor([[0.0, 1.0], [0.0, 1.0]])
    init_weights = torch.tensor([0, 0])
    next_token_ids, ancestors, weights = sampler(logits, init_weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[1], [0]]))
    assert ancestors.equal(torch.tensor([0, 0]))
    assert weights.equal(torch.tensor([logprobs[0][1], logprobs[0][0]]))

    # One beam, batch of two sequences. Reduces to Greedy Search.
    sampler = BeamSearchSampler(1)
    logits = torch.tensor([[0.0, 1.0], [2.0, 0.0]])
    weights = torch.tensor([0, 0])
    next_token_ids, ancestors, weights = sampler(logits, weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[1], [0]]))
    assert ancestors.equal(torch.tensor([0, 1]))
    assert weights.equal(torch.tensor([logprobs[0][1], logprobs[1][0]]))

    # Two beams, batch of two sequences
    sampler = BeamSearchSampler(2)
    logits = torch.tensor([[0.0, 1.0], [2.0, 0.0], [3.0, 2.0], [0.0, 1.0]])
    init_weights = torch.tensor([0, 0, 2.0, 0])
    next_token_ids, ancestors, weights = sampler(logits, init_weights, None)

    logprobs = compute_logprobs(logits)
    assert next_token_ids.equal(torch.tensor([[0], [1], [0], [1]]))
    assert ancestors.equal(torch.tensor([1, 0, 2, 2]))
    assert weights.equal(
        torch.tensor(
            [
                logprobs[1][0],
                logprobs[0][1],
                init_weights[2] + logprobs[2][0],
                init_weights[2] + logprobs[2][1],
            ]
        )
    )
